@@ -3,6 +3,8 @@
 
 #include <gdk/gdktypes.h>
 #include <gdk/gdkgc.h>
+#include <gdk/gdkrgb.h>
+#include <gdk-pixbuf/gdk-pixbuf.h>
 
 #ifdef __cplusplus
 extern "C" {
@@ -130,13 +132,44 @@ struct _GdkDrawableClass
                                           gint         height,
                                           gint        *composite_x_offset,
                                           gint        *composite_y_offset);
-  
-  void         (*_gdk_reserved1) (void);
-  void         (*_gdk_reserved2) (void);
-  void         (*_gdk_reserved3) (void);
-  void         (*_gdk_reserved4) (void);
-  void         (*_gdk_reserved5) (void);
-  void         (*_gdk_reserved6) (void);
+
+  void         (*_draw_pixbuf) (GdkDrawable *drawable,
+			        GdkGC       *gc,
+			        GdkPixbuf   *pixbuf,
+			        gint         src_x,
+			        gint         src_y,
+			        gint         dest_x,
+			        gint         dest_y,
+			        gint         width,
+			        gint         height,
+			        GdkRgbDither dither,
+			        gint         x_dither,
+			        gint         y_dither);
+  GdkImage*    (*_copy_to_image) (GdkDrawable    *drawable,
+				  GdkImage       *image,
+				  gint            src_x,
+				  gint            src_y,
+				  gint            dest_x,
+				  gint            dest_y,
+				  gint            width,
+				  gint            height);
+
+  /* Padding for future expansion */
+  void         (*_gdk_reserved1)  (void);
+  void         (*_gdk_reserved2)  (void);
+  void         (*_gdk_reserved3)  (void);
+  void         (*_gdk_reserved4)  (void);
+  void         (*_gdk_reserved5)  (void);
+  void         (*_gdk_reserved6)  (void);
+  void         (*_gdk_reserved7)  (void);
+  void         (*_gdk_reserved9)  (void);
+  void         (*_gdk_reserved10) (void);
+  void         (*_gdk_reserved11) (void);
+  void         (*_gdk_reserved12) (void);
+  void         (*_gdk_reserved13) (void);
+  void         (*_gdk_reserved14) (void);
+  void         (*_gdk_reserved15) (void);
+  void         (*_gdk_reserved16) (void);
 };
 
 GType           gdk_drawable_get_type     (void);
