@@ -24,6 +24,7 @@
 #include "gtkwindow.h"  /* in lack of GtkAcceleratable */
 
 #include <string.h>
+#include <errno.h>
 #include <fcntl.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -104,18 +105,18 @@ _gtk_accel_path_is_valid (const gchar *accel_path)
 }
 
 /**
- * gtk_accel_map_add_entry
+ * gtk_accel_map_add_entry:
  * @accel_path: valid accelerator path
  * @accel_key:  the accelerator key
  * @accel_mods: the accelerator modifiers
  *
- * Register a new accelerator with the global accelerator map.
+ * Registers a new accelerator with the global accelerator map.
  * This function should only be called once per @accel_path
  * with the canonical @accel_key and @accel_mods for this path.
  * To change the accelerator during runtime programatically, use
  * gtk_accel_map_change_entry().
  * The accelerator path must consist of "&lt;WINDOWTYPE&gt;/Category1/Category2/.../Action",
- * where WINDOWTYPE should be a unique application-specific identifier, that
+ * where &lt;WINDOWTYPE&gt; should be a unique application-specific identifier, that
  * corresponds to the kind of window the accelerator is being used in, e.g. "Gimp-Image",
  * "Abiword-Document" or "Gnumeric-Settings".
  * The Category1/.../Action portion is most appropriately chosen by the action the
@@ -125,9 +126,9 @@ _gtk_accel_path_is_valid (const gchar *accel_path)
  * "&lt;Gimp-Toolbox&gt;/File/Dialogs/Tool Options...".
  */
 void
-gtk_accel_map_add_entry (const gchar *accel_path,
-			 guint        accel_key,
-			 guint        accel_mods)
+gtk_accel_map_add_entry (const gchar    *accel_path,
+			 guint           accel_key,
+			 GdkModifierType accel_mods)
 {
   AccelEntry *entry;
 
@@ -164,9 +165,9 @@ gtk_accel_map_add_entry (const gchar *accel_path,
 }
 
 /**
- * gtk_accel_map_lookup_entry
- * @accel_path:  valid accelerator path
- * @key:         accelerator key to be filled in (optional)
+ * gtk_accel_map_lookup_entry:
+ * @accel_path:  a valid accelerator path
+ * @key:         the accelerator key to be filled in (optional)
  * @returns:     %TRUE if @accel_path is known, %FALSE otherwise
  *
  * Looks up the accelerator entry for @accel_path and fills in @key.
@@ -212,6 +213,10 @@ g_hash_table_slist_values (GHashTable *hash_table)
   return slist;
 }
 
+/* if simulate==TRUE, return whether accel_path can be changed to
+ * accel_key && accel_mods. otherwise, return whether accel_path
+ * was actually changed.
+ */
 static gboolean
 internal_change_entry (const gchar    *accel_path,
 		       guint           accel_key,
@@ -220,7 +225,7 @@ internal_change_entry (const gchar    *accel_path,
 		       gboolean	       simulate)
 {
   GSList *node, *slist, *win_list, *group_list, *replace_list = NULL;
-  GHashTable *group_hm, *win_hm;
+  GHashTable *group_hm, *window_hm;
   gboolean change_accel, removable, can_change = TRUE, seen_accel = FALSE;
   GQuark entry_quark;
   AccelEntry *entry = accel_path_lookup (accel_path);
@@ -241,7 +246,11 @@ internal_change_entry (const gchar    *accel_path,
 
   /* if there's nothing to change, not much todo either */
   if (entry->accel_key == accel_key && entry->accel_mods == accel_mods)
-    return FALSE;
+    {
+      if (!simulate)
+	entry->changed = TRUE;
+      return simulate ? TRUE : FALSE;
+    }
 
   /* nobody's interested, easy going */
   if (!entry->groups)
@@ -258,7 +267,7 @@ internal_change_entry (const gchar    *accel_path,
   /* 1) fetch all accel groups affected by this entry */
   entry_quark = g_quark_try_string (entry->accel_path);
   group_hm = g_hash_table_new (NULL, NULL);
-  win_hm = g_hash_table_new (NULL, NULL);
+  window_hm = g_hash_table_new (NULL, NULL);
   for (slist = entry->groups; slist; slist = slist->next)
     g_hash_table_insert (group_hm, slist->data, slist->data);
 
@@ -269,13 +278,13 @@ internal_change_entry (const gchar    *accel_path,
       GtkAccelGroup *group = slist->data;
 
       for (node = group->acceleratables; node; node = node->next)
-	g_hash_table_insert (win_hm, node->data, node->data);
+	g_hash_table_insert (window_hm, node->data, node->data);
     }
   g_slist_free (group_list);
 
   /* 3) include all accel groups used by acceleratables */
-  win_list = g_hash_table_slist_values (win_hm);
-  g_hash_table_destroy (win_hm);
+  win_list = g_hash_table_slist_values (window_hm);
+  g_hash_table_destroy (window_hm);
   for (slist = win_list; slist; slist = slist->next)
     for (node = gtk_accel_groups_from_object (slist->data); node; node = node->next)
       g_hash_table_insert (group_hm, node->data, node->data);
@@ -370,14 +379,14 @@ internal_change_entry (const gchar    *accel_path,
 }
 
 /**
- * gtk_accel_map_change_entry
- * @accel_path:  valid accelerator path
- * @accel_key:   new accelerator key
- * @accel_mods:  new accelerator modifiers
+ * gtk_accel_map_change_entry:
+ * @accel_path:  a valid accelerator path
+ * @accel_key:   the new accelerator key
+ * @accel_mods:  the new accelerator modifiers
  * @replace:     %TRUE if other accelerators may be deleted upon conflicts
  * @returns:     %TRUE if the accelerator could be changed, %FALSE otherwise
  *
- * Change the @accel_key and @accel_mods currently associated with @accel_path.
+ * Changes the @accel_key and @accel_mods currently associated with @accel_path.
  * Due to conflicts with other accelerators, a change may not always be possible,
  * @replace indicates whether other accelerators may be deleted to resolve such
  * conflicts. A change will only occur if all conflicts could be resolved (which
@@ -449,7 +458,7 @@ accel_map_parse_statement (GScanner *scanner)
     {
       guint (*parser_func) (GScanner*);
 
-      parser_func = scanner->value.v_symbol;
+      parser_func = (guint (*) (GScanner *))scanner->value.v_symbol;
 
       expected_token = parser_func (scanner);
     }
@@ -480,6 +489,12 @@ accel_map_parse_statement (GScanner *scanner)
     }
 }
 
+/**
+ * gtk_accel_map_load_scanner:
+ * @scanner: a #GScanner which has already been provided with an input file
+ *
+ * #GScanner variant of gtk_accel_map_load().
+ */
 void
 gtk_accel_map_load_scanner (GScanner *scanner)
 {
@@ -498,7 +513,8 @@ gtk_accel_map_load_scanner (GScanner *scanner)
   symbol_2_token = scanner->config->symbol_2_token;
   scanner->config->symbol_2_token = FALSE;
   saved_symbol = g_scanner_lookup_symbol (scanner, "gtk_accel_path");
-  g_scanner_scope_add_symbol (scanner, 0, "gtk_accel_path", accel_map_parse_accel_path);
+  g_scanner_scope_add_symbol (scanner, 0, "gtk_accel_path", 
+			      (gpointer) accel_map_parse_accel_path);
 
   /* outer parsing loop
    */
@@ -522,10 +538,11 @@ gtk_accel_map_load_scanner (GScanner *scanner)
 }
 
 /**
- * gtk_accel_map_load_fd
- * @fd: valid readable file descriptor
+ * gtk_accel_map_load_fd:
+ * @fd: a valid readable file descriptor
  *
  * Filedescriptor variant of gtk_accel_map_load().
+ *
  * Note that the file descriptor will not be closed by this function.
  */
 void
@@ -545,7 +562,7 @@ gtk_accel_map_load_fd (gint fd)
 }
 
 /**
- * gtk_accel_map_load
+ * gtk_accel_map_load:
  * @file_name: a file containing accelerator specifications
  *
  * Parses a file previously saved with gtk_accel_map_save() for
@@ -570,15 +587,38 @@ gtk_accel_map_load (const gchar *file_name)
   close (fd);
 }
 
+static gboolean
+write_all (gint   fd,
+	   gchar *buf,
+	   gsize  to_write)
+{
+  while (to_write > 0)
+    {
+      gssize count = write (fd, buf, to_write);
+      if (count < 0)
+	{
+	  if (errno != EINTR)
+	    return FALSE;
+	}
+      else
+	{
+	  to_write -= count;
+	  buf += count;
+	}
+    }
+
+  return TRUE;
+}
+
 static void
 accel_map_print (gpointer        data,
 		 const gchar    *accel_path,
 		 guint           accel_key,
-		 guint           accel_mods,
+		 GdkModifierType accel_mods,
 		 gboolean        changed)
 {
   GString *gstring = g_string_new (changed ? NULL : "; ");
-  gint err, fd = GPOINTER_TO_INT (data);
+  gint fd = GPOINTER_TO_INT (data);
   gchar *tmp, *name;
 
   g_string_append (gstring, "(gtk_accel_path \"");
@@ -597,25 +637,23 @@ accel_map_print (gpointer        data,
 
   g_string_append (gstring, "\")\n");
 
-  do
-    err = write (fd, gstring->str, gstring->len);
-  while (err < 0 && errno == EINTR);
+  write_all (fd, gstring->str, gstring->len);
 
   g_string_free (gstring, TRUE);
 }
 
 /**
- * gtk_accel_map_save_fd
- * @fd: valid writable file descriptor
+ * gtk_accel_map_save_fd:
+ * @fd: a valid writable file descriptor
  *
  * Filedescriptor variant of gtk_accel_map_save().
+ *
  * Note that the file descriptor will not be closed by this function.
  */
 void
 gtk_accel_map_save_fd (gint fd)
 {
   GString *gstring;
-  gint err;
 
   g_return_if_fail (fd >= 0);
 
@@ -626,15 +664,15 @@ gtk_accel_map_save_fd (gint fd)
   g_string_append (gstring, "; this file is an automated accelerator map dump\n");
   g_string_append (gstring, ";\n");
 
-  do
-    err = write (fd, gstring->str, gstring->len);
-  while (err < 0 && errno == EINTR);
+  write_all (fd, gstring->str, gstring->len);
+  
+  g_string_free (gstring, TRUE);
 
   gtk_accel_map_foreach (GINT_TO_POINTER (fd), accel_map_print);
 }
 
 /**
- * gtk_accel_map_save
+ * gtk_accel_map_save:
  * @file_name: the file to contain accelerator specifications
  *
  * Saves current accelerator specifications (accelerator path, key
@@ -659,13 +697,15 @@ gtk_accel_map_save (const gchar *file_name)
 }
 
 /**
- * gtk_accel_map_foreach
+ * gtk_accel_map_foreach:
  * @data:         data to be passed into @foreach_func
- * @foreach_func: function to be executed for each accel map entry
+ * @foreach_func: function to be executed for each accel map entry which
+ *                is not filtered out
  *
- * Loop over the entries in the accelerator map, and execute
- * @foreach_func on each. The signature of @foreach_func is that of
- * #GtkAccelMapForeach, the @changed parameter indicates whether
+ * Loops over the entries in the accelerator map whose accel path 
+ * doesn't match any of the filters added with gtk_accel_map_add_filter(), 
+ * and execute @foreach_func on each. The signature of @foreach_func is 
+ * that of #GtkAccelMapForeach, the @changed parameter indicates whether
  * this accelerator was changed during runtime (thus, would need
  * saving during an accelerator map dump).
  */
@@ -693,6 +733,17 @@ gtk_accel_map_foreach (gpointer           data,
   g_slist_free (entries);
 }
 
+/**
+ * gtk_accel_map_foreach_unfiltered:
+ * @data:         data to be passed into @foreach_func
+ * @foreach_func: function to be executed for each accel map entry
+ *
+ * Loops over all entries in the accelerator map, and execute
+ * @foreach_func on each. The signature of @foreach_func is that of
+ * #GtkAccelMapForeach, the @changed parameter indicates whether
+ * this accelerator was changed during runtime (thus, would need
+ * saving during an accelerator map dump).
+ */
 void
 gtk_accel_map_foreach_unfiltered (gpointer           data,
 				  GtkAccelMapForeach foreach_func)
@@ -712,6 +763,19 @@ gtk_accel_map_foreach_unfiltered (gpointer           data,
   g_slist_free (entries);
 }
 
+/**
+ * gtk_accel_map_add_filter:
+ * @filter_pattern: a pattern (see #GPatternSpec)
+ *
+ * Adds a filter to the global list of accel path filters.
+ *
+ * Accel map entries whose accel path matches one of the filters
+ * are skipped by gtk_accel_map_foreach().
+ *
+ * This function is intended for GTK+ modules that create their own
+ * menus, but don't want them to be saved into the applications accelerator
+ * map dump.
+ */
 void
 gtk_accel_map_add_filter (const gchar *filter_pattern)
 {
