@@ -5,7 +5,7 @@
 
 #ifdef ENABLE_NLS
 #include<libintl.h>
-#define _(String) dgettext("gtk+",String)
+#define _(String) dgettext(GETTEXT_PACKAGE,String)
 #ifdef gettext_noop
 #define N_(String) gettext_noop(String)
 #else
