@@ -430,8 +430,10 @@ gtk_list_store_finalize (GObject *object)
   
   if (list_store->default_sort_destroy)
     {
-      (* list_store->default_sort_destroy) (list_store->default_sort_data);
+      GtkDestroyNotify d = list_store->default_sort_destroy;
+
       list_store->default_sort_destroy = NULL;
+      d (list_store->default_sort_data);
       list_store->default_sort_data = NULL;
     }
 
@@ -514,6 +516,12 @@ gtk_list_store_get_path (GtkTreeModel *tree_model,
 
   g_return_val_if_fail (GTK_IS_LIST_STORE (tree_model), NULL);
   g_return_val_if_fail (iter->stamp == GTK_LIST_STORE (tree_model)->stamp, NULL);
+  if (G_SLIST (iter->user_data) == G_SLIST (GTK_LIST_STORE (tree_model)->tail))
+    {
+      retval = gtk_tree_path_new ();
+      gtk_tree_path_append_index (retval, GTK_LIST_STORE (tree_model)->length - 1);
+      return retval;
+    }
 
   for (list = G_SLIST (GTK_LIST_STORE (tree_model)->root); list; list = list->next)
     {
@@ -646,11 +654,11 @@ static gboolean
 gtk_list_store_real_set_value (GtkListStore *list_store,
 			       GtkTreeIter  *iter,
 			       gint          column,
-			       GValue       *value)
+			       GValue       *value,
+			       gboolean      sort)
 {
   GtkTreeDataList *list;
   GtkTreeDataList *prev;
-  GtkTreePath *path;
   GValue real_value = {0, };
   gboolean converted = FALSE;
   gint orig_column = column;
@@ -690,13 +698,11 @@ gtk_list_store_real_set_value (GtkListStore *list_store,
     {
       if (column == 0)
 	{
-	  path = gtk_list_store_get_path (GTK_TREE_MODEL (list_store), iter);
 	  if (converted)
 	    _gtk_tree_data_list_value_to_node (list, &real_value);
 	  else
 	    _gtk_tree_data_list_value_to_node (list, value);
 	  retval = TRUE;
-	  gtk_tree_path_free (path);
 	  if (converted)
 	    g_value_unset (&real_value);
 	  return retval;
@@ -726,17 +732,15 @@ gtk_list_store_real_set_value (GtkListStore *list_store,
       column --;
     }
 
-  path = gtk_list_store_get_path (GTK_TREE_MODEL (list_store), iter);
   if (converted)
     _gtk_tree_data_list_value_to_node (list, &real_value);
   else
     _gtk_tree_data_list_value_to_node (list, value);
   retval = TRUE;
-  gtk_tree_path_free (path);
   if (converted)
     g_value_unset (&real_value);
 
-  if (GTK_LIST_STORE_IS_SORTED (list_store))
+  if (sort && GTK_LIST_STORE_IS_SORTED (list_store))
     gtk_list_store_sort_iter_changed (list_store, iter, orig_column);
 
   return retval;
@@ -766,7 +770,7 @@ gtk_list_store_set_value (GtkListStore *list_store,
   g_return_if_fail (column >= 0 && column < list_store->n_columns);
   g_return_if_fail (G_IS_VALUE (value));
 
-  if (gtk_list_store_real_set_value (list_store, iter, column, value))
+  if (gtk_list_store_real_set_value (list_store, iter, column, value, TRUE))
     {
       GtkTreePath *path;
 
@@ -793,12 +797,34 @@ gtk_list_store_set_valist (GtkListStore *list_store,
 {
   gint column;
   gboolean emit_signal = FALSE;
+  gboolean maybe_need_sort = FALSE;
+  GtkTreeIterCompareFunc func = NULL;
 
   g_return_if_fail (GTK_IS_LIST_STORE (list_store));
   g_return_if_fail (VALID_ITER (iter, list_store));
 
   column = va_arg (var_args, gint);
 
+  if (GTK_LIST_STORE_IS_SORTED (list_store))
+    {
+      if (list_store->sort_column_id != -1)
+	{
+	  GtkTreeDataSortHeader *header;
+	  header = _gtk_tree_data_list_get_header (list_store->sort_list,
+						   list_store->sort_column_id);
+	  g_return_if_fail (header != NULL);
+	  g_return_if_fail (header->func != NULL);
+	  func = header->func;
+	}
+      else
+	{
+	  func = list_store->default_sort_func;
+	}
+    }
+
+  if (func != gtk_tree_data_list_compare_func)
+    maybe_need_sort = TRUE;
+
   while (column != -1)
     {
       GValue value = { 0, };
@@ -827,13 +853,21 @@ gtk_list_store_set_valist (GtkListStore *list_store,
       emit_signal = gtk_list_store_real_set_value (list_store,
 						   iter,
 						   column,
-						   &value) || emit_signal;
+						   &value,
+						   FALSE) || emit_signal;
+
+      if (func == gtk_tree_data_list_compare_func &&
+	  column == list_store->sort_column_id)
+	maybe_need_sort = TRUE;
 
       g_value_unset (&value);
 
       column = va_arg (var_args, gint);
     }
 
+  if (maybe_need_sort && GTK_LIST_STORE_IS_SORTED (list_store))
+    gtk_list_store_sort_iter_changed (list_store, iter, list_store->sort_column_id);
+
   if (emit_signal)
     {
       GtkTreePath *path;
@@ -931,8 +965,6 @@ gtk_list_store_remove_silently (GtkListStore *list_store,
     if (iter->user_data == list_store->tail)
       list_store->tail = prev;
   }
-
-  list_store->stamp ++;
 }
 
 /**
@@ -1194,7 +1226,7 @@ gtk_list_store_insert_after (GtkListStore *list_store,
   validate_list_store (list_store);
 
   path = gtk_tree_path_new ();
-  gtk_tree_path_append_index (path, i);
+  gtk_tree_path_append_index (path, i + 1);
   gtk_tree_model_row_inserted (GTK_TREE_MODEL (list_store), path, iter);
   gtk_tree_path_free (path);
 }
@@ -1405,13 +1437,20 @@ gtk_list_store_drag_data_received (GtkTreeDragDest   *drag_dest,
                                        prev))
             {
               GtkTreeIter tmp_iter = dest_iter;
-              gtk_list_store_insert_after (GTK_LIST_STORE (tree_model),
-                                           &dest_iter,
-                                           &tmp_iter);
+
+	      if (GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tree_model), "gtk-tree-model-drop-append")))
+		gtk_list_store_append (GTK_LIST_STORE (tree_model), &dest_iter);
+	      else
+		gtk_list_store_insert_after (GTK_LIST_STORE (tree_model),
+					     &dest_iter, &tmp_iter);
+
               retval = TRUE;
             }
         }
 
+      g_object_set_data (G_OBJECT (tree_model), "gtk-tree-model-drop-append",
+			 NULL);
+
       gtk_tree_path_free (prev);
 
       /* If we succeeded in creating dest_iter, copy data from src
@@ -1750,6 +1789,7 @@ gtk_list_store_sort_iter_changed (GtkListStore *list_store,
 
   if ((!list->next) && (cmp_a > 0))
     {
+      new_location++;
       list->next = G_SLIST (iter->user_data);
       list_store->tail = list->next;
     }
@@ -1795,7 +1835,7 @@ gtk_list_store_sort_iter_changed (GtkListStore *list_store,
   tmp_iter.user_data = NULL;
 
   gtk_tree_model_rows_reordered (GTK_TREE_MODEL (list_store),
-				 tmp_path, &tmp_iter,
+				 tmp_path, NULL,
 				 new_order);
 
   gtk_tree_path_free (tmp_path);
@@ -1887,7 +1927,12 @@ gtk_list_store_set_sort_func (GtkTreeSortable        *sortable,
     }
 
   if (header->destroy)
-    (* header->destroy) (header->data);
+    {
+      GtkDestroyNotify d = header->destroy;
+
+      header->destroy = NULL;
+      d (header->data);
+    }
 
   header->func = func;
   header->data = data;
@@ -1906,7 +1951,12 @@ gtk_list_store_set_default_sort_func (GtkTreeSortable        *sortable,
   g_return_if_fail (GTK_IS_LIST_STORE (sortable));
 
   if (list_store->default_sort_destroy)
-    (* list_store->default_sort_destroy) (list_store->default_sort_data);
+    {
+      GtkDestroyNotify d = list_store->default_sort_destroy;
+
+      list_store->default_sort_destroy = NULL;
+      d (list_store->default_sort_data);
+    }
 
   list_store->default_sort_func = func;
   list_store->default_sort_data = data;
