@@ -377,6 +377,31 @@ _gtk_get_module_path (const gchar *type)
   return result;
 }
 
+/* Like g_module_path, but use .la as the suffix
+ */
+static gchar*
+module_build_la_path (const gchar *directory,
+		      const gchar *module_name)
+{
+	gchar *filename;
+	gchar *result;
+	
+	if (strncmp (module_name, "lib", 3) == 0)
+		filename = (gchar *)module_name;
+	else
+		filename =  g_strconcat ("lib", module_name, ".la", NULL);
+
+	if (directory && *directory)
+		result = g_build_filename (directory, filename, NULL);
+	else
+		result = g_strdup (filename);
+
+	if (filename != module_name)
+		g_free (filename);
+
+	return result;
+}
+
 /**
  * _gtk_find_module:
  * @name: the name of the module
@@ -402,15 +427,23 @@ _gtk_find_module (const gchar *name,
   paths = _gtk_get_module_path (type);
   for (path = paths; *path; path++)
     {
-      gchar *tmp_name = g_module_build_path (*path, name);
-	    
+      gchar *tmp_name;
+
+      tmp_name = g_module_build_path (*path, name);
       if (g_file_test (tmp_name, G_FILE_TEST_EXISTS))
 	{
 	  module_name = tmp_name;
 	  goto found;
 	}
-      else
-	g_free(tmp_name);
+      g_free(tmp_name);
+
+      tmp_name = module_build_la_path (*path, name);
+      if (g_file_test (tmp_name, G_FILE_TEST_EXISTS))
+	{
+	  module_name = tmp_name;
+	  goto found;
+	}
+      g_free(tmp_name);
     }
 
   g_strfreev (paths);
@@ -420,8 +453,7 @@ _gtk_find_module (const gchar *name,
 }
 
 static GModule *
-find_module (gchar      **module_path,
-	     const gchar *name)
+find_module (const gchar *name)
 {
   GModule *module;
   gchar *module_name;
@@ -443,7 +475,6 @@ find_module (gchar      **module_path,
 
 static GSList *
 load_module (GSList      *gtk_modules,
-	     gchar      **module_path,
 	     const gchar *name)
 {
   GtkModuleInitFunc modinit_func = NULL;
@@ -451,15 +482,16 @@ load_module (GSList      *gtk_modules,
   
   if (g_module_supported ())
     {
-      module = find_module (module_path, name);
+      module = find_module (name);
       if (module &&
 	  g_module_symbol (module, "gtk_module_init", (gpointer *) &modinit_func) &&
 	  modinit_func)
 	{
-	  if (!g_slist_find (gtk_modules, modinit_func))
+	  if (!g_slist_find (gtk_modules, (gconstpointer) modinit_func))
 	    {
 	      g_module_make_resident (module);
-	      gtk_modules = g_slist_prepend (gtk_modules, modinit_func);
+	      gtk_modules = g_slist_prepend (gtk_modules,
+					     (gpointer) modinit_func);
 	    }
 	  else
 	    {
@@ -483,18 +515,16 @@ load_module (GSList      *gtk_modules,
 static GSList *
 load_modules (const char *module_str)
 {
-  gchar **module_path = get_module_path ();
   gchar **module_names = pango_split_file_list (module_str);
   GSList *gtk_modules = NULL;
   gint i;
   
   for (i = 0; module_names[i]; i++)
-    gtk_modules = load_module (gtk_modules, module_path, module_names[i]);
+    gtk_modules = load_module (gtk_modules, module_names[i]);
   
   gtk_modules = g_slist_reverse (gtk_modules);
   
   g_strfreev (module_names);
-  g_strfreev (module_path);
 
   return gtk_modules;
 }
@@ -729,7 +759,7 @@ gtk_init_check (int	 *argc,
 	{
 	  GtkModuleInitFunc modinit;
 	  
-	  modinit = slist->data;
+	  modinit = (GtkModuleInitFunc) slist->data;
 	  modinit (argc, argv);
 	}
     }
@@ -980,6 +1010,121 @@ gtk_main_iteration_do (gboolean blocking)
     return TRUE;
 }
 
+/* private libgtk to libgdk interfaces
+ */
+gboolean gdk_pointer_grab_info_libgtk_only  (GdkWindow **grab_window,
+					     gboolean   *owner_events);
+gboolean gdk_keyboard_grab_info_libgtk_only (GdkWindow **grab_window,
+					     gboolean   *owner_events);
+
+static void
+rewrite_events_translate (GdkWindow *old_window,
+			  GdkWindow *new_window,
+			  gdouble   *x,
+			  gdouble   *y)
+{
+  gint old_origin_x, old_origin_y;
+  gint new_origin_x, new_origin_y;
+
+  gdk_window_get_origin	(old_window, &old_origin_x, &old_origin_y);
+  gdk_window_get_origin	(new_window, &new_origin_x, &new_origin_y);
+
+  *x += new_origin_x - old_origin_x;
+  *y += new_origin_y - old_origin_y;
+}
+
+static GdkEvent *
+rewrite_event_for_window (GdkEvent  *event,
+			  GdkWindow *new_window)
+{
+  event = gdk_event_copy (event);
+
+  switch (event->type)
+    {
+    case GDK_SCROLL:
+      rewrite_events_translate (event->any.window,
+				new_window,
+				&event->scroll.x, &event->scroll.y);
+      break;
+    case GDK_BUTTON_PRESS:
+    case GDK_2BUTTON_PRESS:
+    case GDK_3BUTTON_PRESS:
+    case GDK_BUTTON_RELEASE:
+      rewrite_events_translate (event->any.window,
+				new_window,
+				&event->button.x, &event->button.y);
+      break;
+    case GDK_MOTION_NOTIFY:
+      rewrite_events_translate (event->any.window,
+				new_window,
+				&event->motion.x, &event->motion.y);
+      break;
+    case GDK_KEY_PRESS:
+    case GDK_KEY_RELEASE:
+    case GDK_PROXIMITY_IN:
+    case GDK_PROXIMITY_OUT:
+      break;
+
+    default:
+      return event;
+    }
+
+  g_object_unref (event->any.window);
+  event->any.window = g_object_ref (new_window);
+
+  return event;
+}
+
+/* If there is a pointer or keyboard grab in effect with owner_events = TRUE,
+ * then what X11 does is deliver the event normally if it was going to this
+ * client, otherwise, delivers it in terms of the grab window. This function
+ * rewrites events to the effect that events going to the same window group
+ * are delivered normally, otherwise, the event is delivered in terms of the
+ * grab window.
+ */
+static GdkEvent *
+rewrite_event_for_grabs (GdkEvent *event)
+{
+  GdkWindow *grab_window;
+  GtkWidget *event_widget, *grab_widget;
+  gboolean owner_events;
+
+  switch (event->type)
+    {
+    case GDK_SCROLL:
+    case GDK_BUTTON_PRESS:
+    case GDK_2BUTTON_PRESS:
+    case GDK_3BUTTON_PRESS:
+    case GDK_BUTTON_RELEASE:
+    case GDK_MOTION_NOTIFY:
+    case GDK_PROXIMITY_IN:
+    case GDK_PROXIMITY_OUT:
+      if (!gdk_pointer_grab_info_libgtk_only (&grab_window, &owner_events) ||
+	  !owner_events)
+	return NULL;
+      break;
+
+    case GDK_KEY_PRESS:
+    case GDK_KEY_RELEASE:
+      if (!gdk_keyboard_grab_info_libgtk_only (&grab_window, &owner_events) ||
+	  !owner_events)
+	return NULL;
+      break;
+
+    default:
+      return NULL;
+    }
+
+  event_widget = gtk_get_event_widget (event);
+  gdk_window_get_user_data (grab_window, (void**) &grab_widget);
+
+  if (grab_widget &&
+      gtk_main_get_window_group (grab_widget) != gtk_main_get_window_group (event_widget))
+    return rewrite_event_for_window (event, grab_window);
+  else
+    return NULL;
+}
+
 void 
 gtk_main_do_event (GdkEvent *event)
 {
@@ -987,6 +1132,7 @@ gtk_main_do_event (GdkEvent *event)
   GtkWidget *grab_widget;
   GtkWindowGroup *window_group;
   GdkEvent *next_event;
+  GdkEvent *rewritten_event = NULL;
   GList *tmp_list;
 
   /* If there are any events pending then get the next one.
@@ -1045,14 +1191,24 @@ gtk_main_do_event (GdkEvent *event)
 
       return;
     }
+
+  /* If pointer or keyboard grabs are in effect, munge the events
+   * so that each window group looks like a separate app.
+   */
+  rewritten_event = rewrite_event_for_grabs (event);
+  if (rewritten_event)
+    {
+      event = rewritten_event;
+      event_widget = gtk_get_event_widget (event);
+    }
   
+  window_group = gtk_main_get_window_group (event_widget);
+
   /* Push the event onto a stack of current events for
    * gtk_current_event_get().
    */
   current_events = g_list_prepend (current_events, event);
 
-  window_group = gtk_main_get_window_group (event_widget);
-  
   /* If there is a grab in effect...
    */
   if (window_group->grabs)
@@ -1198,6 +1354,9 @@ gtk_main_do_event (GdkEvent *event)
   tmp_list = current_events;
   current_events = g_list_remove_link (current_events, tmp_list);
   g_list_free_1 (tmp_list);
+
+  if (rewritten_event)
+    gdk_event_free (rewritten_event);
 }
 
 gboolean
