@@ -63,6 +63,7 @@ struct _GtkMenuItem
   guint submenu_placement : 1;
   guint submenu_direction : 1;
   guint right_justify: 1;
+  guint timer_from_keypress : 1;
   guint timer;
 };
 
@@ -84,6 +85,12 @@ struct _GtkMenuItemClass
 				 gint        *requisition);
   void (* toggle_size_allocate) (GtkMenuItem *menu_item,
 				 gint         allocation);
+
+  /* Padding for future expansion */
+  void (*_gtk_reserved1) (void);
+  void (*_gtk_reserved2) (void);
+  void (*_gtk_reserved3) (void);
+  void (*_gtk_reserved4) (void);
 };
 
 
