@@ -49,7 +49,7 @@ static void gtk_tooltips_class_init        (GtkTooltipsClass *klass);
 static void gtk_tooltips_init              (GtkTooltips      *tooltips);
 static void gtk_tooltips_destroy           (GtkObject        *object);
 
-static gint gtk_tooltips_event_handler     (GtkWidget   *widget,
+static void gtk_tooltips_event_handler     (GtkWidget   *widget,
                                             GdkEvent    *event);
 static void gtk_tooltips_widget_unmap      (GtkWidget   *widget,
                                             gpointer     data);
@@ -62,6 +62,8 @@ static gint gtk_tooltips_timeout           (gpointer     data);
 static gint gtk_tooltips_paint_window      (GtkTooltips *tooltips);
 static void gtk_tooltips_draw_tips         (GtkTooltips *tooltips);
 
+static gboolean get_keyboard_mode          (GtkWidget   *widget);
+
 static GtkObjectClass *parent_class;
 static const gchar  *tooltips_data_key = "_GtkTooltipsData";
 
@@ -280,7 +282,7 @@ gtk_tooltips_set_tip (GtkTooltips *tooltips,
 
       tooltips->tips_data_list = g_list_append (tooltips->tips_data_list,
                                                 tooltipsdata);
-      gtk_signal_connect_after (GTK_OBJECT (widget), "event",
+      gtk_signal_connect_after (GTK_OBJECT (widget), "event-after",
                                (GtkSignalFunc) gtk_tooltips_event_handler,
                                tooltipsdata);
 
@@ -313,13 +315,14 @@ gtk_tooltips_paint_window (GtkTooltips *tooltips)
 }
 
 static void
-gtk_tooltips_draw_tips (GtkTooltips * tooltips)
+gtk_tooltips_draw_tips (GtkTooltips *tooltips)
 {
   GtkRequisition requisition;
   GtkWidget *widget;
   GtkStyle *style;
   gint x, y, w, h, scr_w, scr_h;
   GtkTooltipsData *data;
+  gboolean keyboard_mode;
 
   if (!tooltips->tip_window)
     gtk_tooltips_force_window (tooltips);
@@ -331,6 +334,8 @@ gtk_tooltips_draw_tips (GtkTooltips * tooltips)
   
   widget = tooltips->active_tips_data->widget;
 
+  keyboard_mode = get_keyboard_mode (widget);
+
   scr_w = gdk_screen_width ();
   scr_h = gdk_screen_height ();
 
@@ -342,10 +347,17 @@ gtk_tooltips_draw_tips (GtkTooltips * tooltips)
   w = requisition.width;
   h = requisition.height;
 
-  gdk_window_get_pointer (NULL, &x, NULL, NULL);
-  gdk_window_get_origin (widget->window, NULL, &y);
+  gdk_window_get_origin (widget->window, &x, &y);
   if (GTK_WIDGET_NO_WINDOW (widget))
-    y += widget->allocation.y;
+    {
+      x += widget->allocation.x;
+      y += widget->allocation.y;
+    }
+
+  x += widget->allocation.width / 2;
+    
+  if (!keyboard_mode)
+    gdk_window_get_pointer (NULL, &x, NULL, NULL);
 
   x -= (w / 2 + 4);
 
@@ -421,6 +433,35 @@ gtk_tooltips_set_active_widget (GtkTooltips *tooltips,
     }
 }
 
+static void
+gtk_tooltips_show_tip (GtkWidget *widget)
+{
+  GtkTooltipsData *tooltipsdata;
+
+  tooltipsdata = gtk_tooltips_data_get (widget);
+
+  if (tooltipsdata &&
+      (!tooltipsdata->tooltips->active_tips_data ||
+       tooltipsdata->tooltips->active_tips_data->widget != widget))
+    {
+      gtk_tooltips_set_active_widget (tooltipsdata->tooltips, widget);
+      gtk_tooltips_draw_tips (tooltipsdata->tooltips);
+    }
+}
+
+static void
+gtk_tooltips_hide_tip (GtkWidget *widget)
+{
+  GtkTooltipsData *tooltipsdata;
+
+  tooltipsdata = gtk_tooltips_data_get (widget);
+
+  if (tooltipsdata &&
+      (tooltipsdata->tooltips->active_tips_data &&
+       tooltipsdata->tooltips->active_tips_data->widget == widget))
+    gtk_tooltips_set_active_widget (tooltipsdata->tooltips, NULL);
+}
+
 static gboolean
 gtk_tooltips_recently_shown (GtkTooltips *tooltips)
 {
@@ -433,70 +474,134 @@ gtk_tooltips_recently_shown (GtkTooltips *tooltips)
   return (msec < STICKY_REVERT_DELAY);
 }
 
-static gint
+static gboolean
+get_keyboard_mode (GtkWidget *widget)
+{
+  GtkWidget *toplevel = gtk_widget_get_toplevel (widget);
+  if (GTK_IS_WINDOW (toplevel))
+    return GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (toplevel), "gtk-tooltips-keyboard-mode"));
+  else
+    return FALSE;
+}
+
+static void
+start_keyboard_mode (GtkWidget *widget)
+{
+  GtkWidget *toplevel = gtk_widget_get_toplevel (widget);
+  if (GTK_IS_WINDOW (toplevel))
+    {
+      GtkWidget *focus = GTK_WINDOW (toplevel)->focus_widget;
+
+      g_object_set_data (G_OBJECT (toplevel), "gtk-tooltips-keyboard-mode", GUINT_TO_POINTER (TRUE));
+
+      if (focus)
+	gtk_tooltips_show_tip (focus);
+    }
+}
+
+static void
+stop_keyboard_mode (GtkWidget *widget)
+{
+  GtkWidget *toplevel = gtk_widget_get_toplevel (widget);
+  if (GTK_IS_WINDOW (toplevel))
+    {
+      GtkWidget *focus = GTK_WINDOW (toplevel)->focus_widget;
+      if (focus)
+	gtk_tooltips_hide_tip (focus);
+      
+      g_object_set_data (G_OBJECT (toplevel), "gtk-tooltips-keyboard-mode", GUINT_TO_POINTER (FALSE));
+    }
+}
+
+static void
 gtk_tooltips_event_handler (GtkWidget *widget,
                             GdkEvent  *event)
 {
   GtkTooltips *tooltips;
   GtkTooltipsData *old_tips_data;
   GtkWidget *event_widget;
+  gboolean keyboard_mode = get_keyboard_mode (widget);
 
   if ((event->type == GDK_LEAVE_NOTIFY || event->type == GDK_ENTER_NOTIFY) &&
       event->crossing.detail == GDK_NOTIFY_INFERIOR)
-    return FALSE;
+    return;
 
-  event_widget = gtk_get_event_widget (event);
-  if (event_widget != widget)
-    return FALSE;
-  
   old_tips_data = gtk_tooltips_data_get (widget);
   tooltips = old_tips_data->tooltips;
 
-  switch (event->type)
+  if (keyboard_mode)
     {
-    case GDK_MOTION_NOTIFY:
-    case GDK_EXPOSE:
-      /* do nothing */
-      break;
-      
-    case GDK_ENTER_NOTIFY:
-      old_tips_data = tooltips->active_tips_data;
-      if (tooltips->enabled &&
-	  (!old_tips_data || old_tips_data->widget != widget))
+      switch (event->type)
 	{
-	  guint delay;
-	  
-	  gtk_tooltips_set_active_widget (tooltips, widget);
-	  
-	  if (tooltips->use_sticky_delay  &&
-	      gtk_tooltips_recently_shown (tooltips))
-	    delay = STICKY_DELAY;
+	case GDK_FOCUS_CHANGE:
+	  if (event->focus_change.in)
+	    gtk_tooltips_show_tip (widget);
 	  else
-	    delay = tooltips->delay;
-	  tooltips->timer_tag = gtk_timeout_add (delay,
-						 gtk_tooltips_timeout,
-						 (gpointer) tooltips);
+	    gtk_tooltips_hide_tip (widget);
+	  break;
+	default:
+	  break;
 	}
-      break;
-
-    case GDK_LEAVE_NOTIFY:
-      {
-	gboolean use_sticky_delay;
-
-	use_sticky_delay = tooltips->tip_window &&
-		GTK_WIDGET_VISIBLE (tooltips->tip_window);
-	gtk_tooltips_set_active_widget (tooltips, NULL);
-	tooltips->use_sticky_delay = use_sticky_delay;
-      }
-      break;
-      
-    default:
-      gtk_tooltips_set_active_widget (tooltips, NULL);
-      return FALSE;
-      break;
     }
+  else
+    {
+      if (event->type != GDK_KEY_PRESS && event->type != GDK_KEY_RELEASE)
+	{
+	  event_widget = gtk_get_event_widget (event);
+	  if (event_widget != widget)
+	    return;
+	}
+  
+      switch (event->type)
+	{
+	case GDK_MOTION_NOTIFY:
+	case GDK_EXPOSE:
+	  /* do nothing */
+	  break;
+	  
+	case GDK_ENTER_NOTIFY:
+	  old_tips_data = tooltips->active_tips_data;
+	  if (tooltips->enabled &&
+	      (!old_tips_data || old_tips_data->widget != widget))
+	    {
+	      guint delay;
+	      
+	      gtk_tooltips_set_active_widget (tooltips, widget);
+	      
+	      if (tooltips->use_sticky_delay  &&
+	      gtk_tooltips_recently_shown (tooltips))
+		delay = STICKY_DELAY;
+	      else
+		delay = tooltips->delay;
+	      tooltips->timer_tag = gtk_timeout_add (delay,
+						     gtk_tooltips_timeout,
+						     (gpointer) tooltips);
+	    }
+	  break;
+	  
+	case GDK_LEAVE_NOTIFY:
+	  {
+	    gboolean use_sticky_delay;
+	    
+	    use_sticky_delay = tooltips->tip_window &&
+	      GTK_WIDGET_VISIBLE (tooltips->tip_window);
+	    gtk_tooltips_set_active_widget (tooltips, NULL);
+	    tooltips->use_sticky_delay = use_sticky_delay;
+	  }
+	  break;
 
-  return FALSE;
+	case GDK_BUTTON_PRESS:
+	case GDK_BUTTON_RELEASE:
+	case GDK_KEY_PRESS:
+	case GDK_KEY_RELEASE:
+	case GDK_PROXIMITY_IN:
+	case GDK_SCROLL:
+	  gtk_tooltips_set_active_widget (tooltips, NULL);
+	  break;
+	default:
+	  break;
+	}
+    }
 }
 
 static void
@@ -525,23 +630,11 @@ gtk_tooltips_widget_remove (GtkWidget *widget,
 }
 
 void
-_gtk_tooltips_show_tip (GtkWidget *widget)
+_gtk_tooltips_toggle_keyboard_mode (GtkWidget *widget)
 {
-  /* Showing the tip from the keyboard */
-
-  /* FIXME this function is completely broken right now,
-   * popdown doesn't occur when it should.
-   */
-  
-  GtkTooltipsData *tooltipsdata;
-
-  tooltipsdata = gtk_tooltips_data_get (widget);
-
-  if (tooltipsdata == NULL)
-    return;
-
-  gtk_tooltips_set_active_widget (tooltipsdata->tooltips,
-                                  widget);
-
-  gtk_tooltips_timeout (tooltipsdata->tooltips);
+  if (get_keyboard_mode (widget))
+    stop_keyboard_mode (widget);
+  else
+    start_keyboard_mode (widget);
 }
+
